#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <glib.h>

// pkg-config --cflags glib-2.0 --libs


#define	consoleOutput 0
#define fileOutput 0
#define debugOutput 0

#define emptyField 0
#define antField 1
#define flavorField 80
#define resourceField 90
#define bauField 100

#define MaxAnts 2000
#define SimulationLength 4000
#define ResourcesPerField 5 // defines how much sugar is on one resource field
#define MultiOperation 500

#define	TASKS 12
#define MPI_CreateAnt 100
#define MPI_MoveAnts 200

#define MPI_AddFlavor 700
#define MPI_RemoveFlavor 701
#define MPI_IncrementResources 702
#define MPI_ResourcePosition 703
#define MPI_TryGatherResource 704
#define MPI_ResourceDepleted 705
#define MPI_IncrementResource 706

#define MPI_SlaveReady 800


struct AntsBau
{
	int AtRow;	// row pos.
	int AtCol;  // column pos.
	int GatheredResources;
};

// global vars
struct AntsBau Bau; // initialized for all threads
/* map is used in thread 0 for debugOutput, 
the other threads do not use this global var
*/

struct _GHashTable hashTbl;

#include "FlavorList.c"
#include "VectorList.c"
#include "AntList.c"
#include "Helpers.c"

/*
Sequentiell:
Laufzeiten Linux mit consoleOutput = false, 
Sekunden /  maxRows / maxCols  / MaxAnts / SimulationLength 
55		 /  240	 / 500		/ 30.000	 / 15.000


Parallel:
Laufzeiten Linux mit consoleOutput = false, 
Sekunden / maxRows / maxCols  / MaxAnts / SimulationLength / Threads

11				 240 	   500		  500		1.000			 / 3
7				 240 	   500		  500		1.000			 / 4
8				 240 	   500		  500		1.000			 / 5

Nach MPI Optimierung und getCoord inline Optimierung

11				 240 	   500		  500		1.000			 / 3
7				 240 	   500		  500		1.000			 / 4
4				 240 	   500		  500		1.000			 / 5
2				 240 	   500		  500		1.000			 / 10
1				 240 	   500		  500		1.000			 / 12
2-3				 240 	   500		  500		1.000			 / 20

8				 240 	   500		  1.000		2.000			 / 12
42				 240 	   500		  2.000		4.000			 / 12
*/



int antIsAtBau(struct Ant a)
{
	return a.AtPos.Row == Bau.AtRow && a.AtPos.Col == Bau.AtCol;
}

void initBau()
{
	Bau.AtRow = maxRows / 2;
	Bau.AtCol = maxCols / 2;
	Bau.GatheredResources = 0;
}

void giveDirectionsToBau(struct Ant *a)
{
	struct Vector2 vec = getCoordinate(a->AtPos.Row, a->AtPos.Col, Bau.AtRow, Bau.AtCol);
	giveDirections(a, vec);
}

int addFlavor(struct ListFlavor *flavorList, int row, int col) {

	struct Vector2 vectorToBau = getCoordinate(row, col, Bau.AtRow, Bau.AtCol);
	int distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col);
	
	struct Flavor fl;
	fl.Row = row;
	fl.Col = col;
	fl.Duration = distanceFromBau*3;
	return appendFlavor(flavorList, fl);
}

void initResource(struct VectorList *resourceList)
{	
	// determine the sugar positions
	int sugarsToCreate = maxRows * maxCols / 300; // exactly 4 sugar fields for 24 rows * 50 cols = 1200 fields	
	int row, col;
	srand (time(NULL));
	for(sugarsToCreate; sugarsToCreate > 0; sugarsToCreate--)
	{		
		int distanceFromBau;
		struct Vector2 sugarPos;
		do
		{
			sugarPos.Row = rand() % maxRows;
			sugarPos.Col = rand() % maxCols;
			struct Vector2 vectorToBau = getCoordinate(sugarPos.Row, sugarPos.Col, Bau.AtRow, Bau.AtCol);
			distanceFromBau = abs(vectorToBau.Row) + abs(vectorToBau.Col); 
		}
		while(distanceFromBau < 10);
		
		for(row = sugarPos.Row; row < sugarPos.Row + 2; row++)
			for(col = sugarPos.Col; col < sugarPos.Col + 3; col++)
			{
				struct Vector2 tmp;
				tmp.IsEmpty = 0;
				tmp.Resources = ResourcesPerField;
				tmp.Row = row;
				tmp.Col = col;
				if(rowAndColWithinBounds(row, col))
					appendVector2(resourceList, tmp);
			}
	}
}

void sendResourcesToSlaves(struct VectorList *resourceList, int tasks)
{
	int i;
	for(i=2; i < tasks; i++)
	{
		struct VectorList *currentResource = resourceList;
		while(currentResource->Next != NULL)
		{
			currentResource = currentResource->Next;
			int resPos[2] = { currentResource->Value.Row, currentResource->Value.Col };
			if(debugOutput && 0) 
				printf("send resource at pos %d %d to task %d\n", resPos[0], resPos[1], i);
			MPI_Send(resPos, 2, MPI_INT, i, MPI_ResourcePosition, MPI_COMM_WORLD);
		}
	}
}


inline void move(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(curPos->Row + movePos->Row >= maxRows)
		movePos->Row = maxRows - curPos->Row - 1;
	if(curPos->Col + movePos->Col >= maxCols)
		movePos->Col = maxCols - curPos->Col - 1;
	if(curPos->Row + movePos->Row < 0)
		movePos->Row = -curPos->Row;
	if(curPos->Col + movePos->Col < 0)
		movePos->Col = -curPos->Col;

	int moveRow = 0, moveCol = 0;
	if( movePos->Row != 0 || movePos->Col != 0)
	{
		if(abs(movePos->Row) > abs(movePos->Col))
			moveRow = abs(movePos->Row) / movePos->Row;
		else
			moveCol = abs(movePos->Col) / movePos->Col;
		movePos->Row -= moveRow;
		movePos->Col -= moveCol;
	}

	curPos->Row = curPos->Row+moveRow;
	curPos->Col = curPos->Col+moveCol;
}

inline void moveRnd(struct Vector2 *curPos, struct Vector2 *movePos)
{
	if(movePos->Row == 0 && movePos->Col == 0) // give new direction
	{
		movePos->Row = rand() % 11 - 5;
		movePos->Col = rand() % 11 - 5;
	}
	move(curPos, movePos); // do 1 step into direction
}

void updateFlavorList(struct ListFlavor *flavorList)
{
	struct ListFlavor *currentFlavor = flavorList;
	while(currentFlavor->Next != NULL)
	{	
		struct ListFlavor *lastFlavor = currentFlavor;
		currentFlavor = currentFlavor->Next;
		if(currentFlavor->Value.Duration > 0)
			currentFlavor->Value.Duration--;
		if(currentFlavor->Value.Duration == 0)
		{
			int flavorPos[2] = { currentFlavor->Value.Row, currentFlavor->Value.Col };
			MPI_Send(flavorPos, 2, MPI_INT, 1, MPI_RemoveFlavor, MPI_COMM_WORLD);
			lastFlavor->Next = currentFlavor->Next;
		}
	}
}

inline void calculateAntMovement(struct Ant *a, struct VectorList *resourceList, struct ListFlavor *flavorList)
{	
	MPI_Status status;
	int antOnResourceFlag = 0;
	if(!a->HasResource)
		antOnResourceFlag = antOnResource(resourceList, a->AtPos.Row, a->AtPos.Col);
	int curPos[2] = { a->AtPos.Row, a->AtPos.Col };
	
	struct Vector2 resPos;
	resPos.IsEmpty = 1;
	if(!a->HasResource && !antOnResourceFlag)
		resPos = antSeesRessource(resourceList, a->AtPos.Row, a->AtPos.Col, a->ViewDist); // row, col, dist, objectId....
	struct Vector2 flavPos;
	flavPos.IsEmpty = 1;
	if(!antOnResourceFlag)
		if(!a->IsFollowingFlavor || (a->MovePos.Row == 0 && a->MovePos.Col == 0))
			flavPos = antSeesFlavor(flavorList, a->AtPos.Row, a->AtPos.Col, a->ViewDist);
	if(a->HasResource && !antOnResourceFlag) {
		addFlavor(flavorList, a->AtPos.Row, a->AtPos.Col);
		MPI_Send(curPos, 2, MPI_INT, 0, MPI_AddFlavor, MPI_COMM_WORLD);
	}
	if(!a->HasResource && antOnResourceFlag)
	{
		MPI_Send(curPos, 2, MPI_INT, 0, MPI_TryGatherResource, MPI_COMM_WORLD);
		MPI_Probe(0, MPI_TryGatherResource, MPI_COMM_WORLD, &status);  //  int source, int tag, MPI_Comm comm, MPI_Status* status)
		
		int b = 0;
		MPI_Recv(&b, 1, MPI_INT, 0, MPI_TryGatherResource, MPI_COMM_WORLD, &status);			
		if(b) {
			a->HasResource = 1;		
			giveDirectionsToBau(a);
		}
		else { // resource depleted
			removeResource(resourceList, a->AtPos.Row, a->AtPos.Col);
		}
	}	
	else if(!a->HasResource && !resPos.IsEmpty)
	{
		struct Vector2 movement;		
		movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, resPos.Row, resPos.Col);	
		giveDirections(a, movement);
		move(&a->AtPos, &a->MovePos);
	}		
	else if(!a->HasResource && a->Caste == 0 && !flavPos.IsEmpty) {
		struct Vector2 movement;
		movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, flavPos.Row, flavPos.Col);	
		giveDirections(a, movement);
		move(&a->AtPos, &a->MovePos);
		a->IsFollowingFlavor = 1;
	}
	else if(a->HasResource && antIsAtBau(*a))
	{
		a->HasResource = 0;
		a->IsFollowingFlavor = 0;
		MPI_Send(&a->AntId, 1, MPI_INT, 0, MPI_IncrementResource, MPI_COMM_WORLD);
	}	
	else
	{
		moveRnd(&a->AtPos, &a->MovePos);
	}	
}


MPI_Status receiveIndexOrVector(int *nbr, struct Vector2 *vec)
{
	MPI_Status status;
	MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	
	int nelements;
	MPI_Get_count(&status, MPI_INT, &nelements);
	
	if(nelements == 1)
	{
		int antIndex;
		MPI_Recv(&antIndex, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		*nbr = antIndex;
	}
	else if (nelements == 2)
	{
		int pos[2];
		MPI_Recv(pos, 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		vec->Row = pos[0];
		vec->Col = pos[1];
		if(pos[0] != -1 && pos[1] != -1)
			vec->IsEmpty = 0;
		else
			vec->IsEmpty = 1;
	}
	else
	{
		if(debugOutput)					
			printf("nelements: %d\n", nelements);
	}
	return status;
}

void master(int ntasks)
{
	int map[maxRows][maxCols]; 
	void clearMap()
	{
		int i, j;	
		for(i = 0; i < maxRows; i++)
			for(j = 0; j < maxCols; j++)
				if(map[i][j] != emptyField)
					map[i][j] = emptyField;
	}
	
	void printMap(int antsOut)
	{
		int i, j;
		for(i = 0; i < maxRows; i++)
		{
			for(j = 0; j < maxCols; j++)
			{		
				int value = map[i][j];
				if(value == bauField)
				{
					printf("b");
				}
				else if(value == resourceField)
				{
					printf("z");
				}
				
				else if(value == emptyField)
				{
					printf(" ");
				}
				else if(value == antField)
				{
					struct Vector2 pos;
					pos.IsEmpty = 0;
					pos.Row = i;
					pos.Col = j;
					/*
					int antsAtPos = antsAtPosition(pos);
					if(antsAtPos > 9)
						printf("+");
					else
					*/
						printf("%d", antField);
				}
				else if(value == flavorField)
				{
					printf("D");
				}
				else
				{
					printf("%d", value);
				}
			}
			printf("\n");
		}
			
		/*
		int totalWorkers = 0;
		int totalScouts = 0;
		for(i = 0; i < antsOut; i++)
		{
			if(Ants[i].Caste == 0)
				totalWorkers++;
			else // == 1
				totalScouts++;
		}
		printf("wokers/scouts: %d/%d resources: %d, ", totalWorkers, totalScouts, Bau.GatheredResources);
		*/
	} // end printMap

	void initMap(struct VectorList *resourceList, struct ListFlavor *flavorList, int currentRound)
	{
		struct ListFlavor *currentFlavor = flavorList;
		while(currentFlavor->Next != NULL)
		{
			currentFlavor = currentFlavor->Next;
			struct Flavor flavorPos = currentFlavor->Value;
			if(flavorPos.Duration > 0)
				map[flavorPos.Row][flavorPos.Col] = flavorField;			
		}
		
		int i, j;
		// draw some sugar
		struct VectorList *currentResource = resourceList;
		while(currentResource->Next != NULL)
		{
			currentResource = currentResource->Next;
			struct Vector2 sugarPos = currentResource->Value;
			map[sugarPos.Row][sugarPos.Col] = resourceField;
		}
		
		map[Bau.AtRow][Bau.AtCol] = bauField;
	}
	// actual main start:
	time_t start;
	time(&start);

	struct VectorList *resourceList;
	resourceList = mallocVector2List();
	initResource(resourceList);
	
	sendResourcesToSlaves(resourceList, ntasks);
	
	struct ListFlavor *flavorList;
	flavorList = mallocFlavorList();	
	
	int antsOut = 0, duration, index;	
	for(duration = 0; duration < SimulationLength; duration++)
	{	
		int currentTask;
		if(consoleOutput)
			clearMap();
		if(antsOut < MaxAnts)
		{
			if(debugOutput)					
				printf("Round %d; Master_CreateAnt\n", duration);
			int newAntCount = MaxAnts / (SimulationLength / 2);
			if(newAntCount == 0)
				newAntCount = 1;
			
			for(newAntCount; newAntCount > 0; newAntCount--)
			{
				currentTask = (antsOut % (TASKS - 2)) + 2;
				MPI_Send(&antsOut, 1, MPI_INT, currentTask, MPI_CreateAnt, MPI_COMM_WORLD);
				// printf("create: current worker task is %d\n", currentTask);
				antsOut++;
			}
		}
		
		for(index = 2; index < ntasks; index++)
		{
			int id;
			MPI_Send(&id, 1, MPI_INT, index, MPI_MoveAnts, MPI_COMM_WORLD);
		}
				
		int threadsBusy = ntasks-2;
		int antInformation[MultiOperation];
		while(threadsBusy != 0) // master
		{
			int antIndex = -1;
			struct Vector2 curVector;			
			//MPI_Status status = receiveIndexOrVector(&antIndex, &curVector);
			MPI_Status status;
			MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
			
			int nelements;
			MPI_Get_count(&status, MPI_INT, &nelements);
			
			if(nelements == 1 && status.MPI_TAG != MPI_MoveAnts)
			{
				int antIndex;
				MPI_Recv(&antIndex, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
			}
			else if (nelements == 2 && status.MPI_TAG != MPI_MoveAnts)
			{
				int pos[2];
				MPI_Recv(pos, 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
				curVector.Row = pos[0];
				curVector.Col = pos[1];
				if(pos[0] != -1 && pos[1] != -1)
					curVector.IsEmpty = 0;
				else
					curVector.IsEmpty = 1;
			}
			
			if(debugOutput && 0)
				printf("Master_Listening\n");
		
			if(status.MPI_TAG == MPI_SlaveReady)
			{
				threadsBusy--; // received all new pos
			}
			else if(status.MPI_TAG == MPI_MoveAnts)
			{
			
				if(debugOutput)
					printf("Master_MoveAnt recv:%d of data\n", nelements);
				MPI_Recv(antInformation, nelements, MPI_INT, MPI_ANY_SOURCE, MPI_MoveAnts, MPI_COMM_WORLD, &status);
				
				for(index = 0; index < nelements; index += 2)
				{
					curVector.Row = antInformation[index];
					curVector.Col = antInformation[index+1];
					map[curVector.Row][curVector.Col] = antField;
					
					if(debugOutput)
					{
						if(curVector.Row < 0 || curVector.Row >= maxRows)
							printf("WARNING ant out of map: at pos %d %d\n", curVector.Row, curVector.Col);
						else if(curVector.Col < 0 || curVector.Col >= maxCols)
							printf("WARNING ant out of map: at pos %d %d\n", curVector.Row, curVector.Col);
						else
							printf("Master_NewPos %d %d\n", curVector.Row, curVector.Col);
					}
				}
			}
			else if(status.MPI_TAG == MPI_AddFlavor)
			{			
				int created = addFlavor(flavorList, curVector.Row, curVector.Col);
				if(created) // the other threads should know
				{
					int flPos[2] = { curVector.Row, curVector.Col };
					for(index = 2; index < ntasks; index++)
					{
						if(index != status.MPI_SOURCE)
							MPI_Send(flPos, 2, MPI_INT, index, MPI_AddFlavor, MPI_COMM_WORLD);
					}
				}
			}
			else if(status.MPI_TAG == MPI_TryGatherResource)
			{
				int b = gatherResource(resourceList, &curVector);
				MPI_Send(&b, 1, MPI_INT, status.MPI_SOURCE, MPI_TryGatherResource, MPI_COMM_WORLD);
				
				if(!b)
				{
					int resPos[2] = { curVector.Row, curVector.Col };
					for(index = 2; index < ntasks; index++)
					{
						if(index != status.MPI_SOURCE)
							MPI_Send(resPos, 2, MPI_INT, index, MPI_ResourceDepleted, MPI_COMM_WORLD);
					}
				}
			}			
			else if(status.MPI_TAG == MPI_IncrementResource)
			{
				Bau.GatheredResources++;
			}
			else if(debugOutput)
			{				
				printf("error on master: %d\n", status.MPI_TAG);
				exit(1);
			}
		}
				
		if(consoleOutput)
		{
			system("clear");
			initMap(resourceList, flavorList, duration);
			printMap(antsOut);			
			printf("resources: %d, ", Bau.GatheredResources);
			printf("Round: %d/%d, ", duration+1, SimulationLength);		
		}
		
		updateFlavorList(flavorList);
		// getchar();	
	} // end for(duration < SimulationLength)

	int timeDiff = (int)time(NULL) - (int)start;
	printf("resources: %d, ", Bau.GatheredResources);
	printf("%d Ameisen in %d seconds.\n", antsOut, timeDiff);
	
	int quit = 999;
	for(index = 1; index < ntasks; index++)
	{
		MPI_Send(&quit, 1, MPI_INT, index, 666, MPI_COMM_WORLD);
	}
	
	freeVector2(resourceList);
	freeFlavor(flavorList);
	
} // end master

void resourceAndFlavorSlave(int tasks)
{
	int map[maxRows][maxCols]; 
	// returns the position of a resource in the ants distance
	
	int running = 1;
	while(running == 1) {
				
		int antIndex;
		struct Vector2 curVec = { -1, -1 };
		MPI_Status status = receiveIndexOrVector(&antIndex, &curVec);
		
		if(debugOutput && 0)					
			printf("Slave_One_Listening...\n");			
				
		if(status.MPI_TAG == 666) {
			running = 0;				
			printf("Thread 1 terminiert nun!\n");
		}
		else if(status.MPI_TAG == MPI_RemoveFlavor)
		{
			int i;
			int flavorPos[2] = { curVec.Row, curVec.Col };
			for(i=2; i < tasks; i++)
			{
				MPI_Send(flavorPos, 2, MPI_INT, i, MPI_RemoveFlavor, MPI_COMM_WORLD);
			}
		}
		else if(debugOutput)
		{				
			printf("error on slave one: %d\n", status.MPI_TAG);
			exit(1);
		}
	}
	
	// regular end of resourceAndFlavorSlave
	
	
	
} // end of resourceAndFlavorSlave

void slave(int rank)
{	
	struct AntList *antList = mallocAntList();
	int antListCount = 0;
	MPI_Status status;
	int running = 1;
	
	struct VectorList *resourceList;
	resourceList = mallocVector2List();
	
	struct ListFlavor *flavorList;
	flavorList = mallocFlavorList();	
	
	srand(time(NULL)+(rank*100)); // every slave needs its own seed for rand
	while(running == 1) {
		MPI_Probe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);  //  int source, int tag, MPI_Comm comm, MPI_Status* status)
		
		if(debugOutput && 0)
			printf("Slave_Listening...\n");			
				
		int count;
		struct Vector2 curVec = { -1, -1 };
		status = receiveIndexOrVector(&count, &curVec);
		
		if(status.MPI_TAG == 666) {
			running = 0;				
			printf("Thread %i terminiert nun!\n", rank);
		}
		else if(status.MPI_TAG == MPI_ResourcePosition) // Resource is saved for worker thread
		{
			struct Vector2 tmp;
			tmp.IsEmpty = 0;
			tmp.Row = curVec.Row;
			tmp.Col = curVec.Col;
			appendVector2(resourceList, tmp);
			if(debugOutput && 0)
				printf("Worker Thread %d: MPI_ResourcePosition at %d %d\n", rank, curVec.Row, curVec.Col);
		}
		else if(status.MPI_TAG == MPI_ResourceDepleted)
		{
			removeResource(resourceList, curVec.Row, curVec.Col);
			if(debugOutput)
				printf("Worker Thread %d: MPI_ResourceDepleted at %d %d\n", rank, curVec.Row, curVec.Col);
		}		
		else if(status.MPI_TAG == MPI_RemoveFlavor)
		{
			removeFlavor(flavorList, curVec.Row, curVec.Col);
			if(debugOutput)
				printf("Helper Thread: MPI_RemoveFlavor at %d %d\n", curVec.Row, curVec.Col);
		}
		else if(status.MPI_TAG == MPI_AddFlavor)
		{
			addFlavor(flavorList, curVec.Row, curVec.Col);
		}		
		else if(status.MPI_TAG == MPI_CreateAnt)
		{
			if(debugOutput)
				printf("Slave_CreateAnt: %d\n", count);
			addAnt(antList, ++antListCount, Bau.AtRow, Bau.AtCol, rank);
	
			// debug stuff 
			/*
			struct Vector2 resPos;
			resPos = resourceList->Next->Value;
			struct Ant *a;
			a = &antList->Next->Value;
			
			
			
			struct Vector2 movement;		
			movement = getCoordinate(a->AtPos.Row, a->AtPos.Col, resPos.Row, resPos.Col);	
			
			giveDirections(a, movement);
			*/
		}
		else if(status.MPI_TAG == MPI_MoveAnts)
		{
			if(debugOutput)
				printf("Worker Thread %d: move ants\n", rank);
		
			int curCounter = 0;
			int antInformation[MultiOperation];
		
			struct AntList *currentAnt = antList;
			while(currentAnt->Next != NULL)
			{
				currentAnt = currentAnt->Next;
				if(debugOutput)
					printf("Slave_MoveAnt id:%d\n", currentAnt->Value.AntId);
				calculateAntMovement(&(currentAnt->Value), resourceList, flavorList);
				antInformation[curCounter] = currentAnt->Value.AtPos.Row;
				antInformation[curCounter+1] = currentAnt->Value.AtPos.Col;
				
				curCounter += 2;
				
				if(curCounter == MultiOperation || currentAnt->Next == NULL)
				{
					if(debugOutput)
						printf("Slave_MoveAnt sending:%d of data\n", curCounter);
					MPI_Send(antInformation, curCounter, MPI_INT, 0, MPI_MoveAnts, MPI_COMM_WORLD);
					curCounter = 0;
				}
			}
			
			MPI_Send(&curCounter, 1, MPI_INT, 0, MPI_SlaveReady, MPI_COMM_WORLD);
		}
		else if(debugOutput)
		{				
			printf("error on slave: %d\n", status.MPI_TAG);
			exit(1);
		}
		
	}
}
guint *hash;

int main(argc, argv)
     int argc;
     char *argv[];
{

	hash = g_int_hash(*gconstpointer);
	hashTbl = g_hash_table_new(hash, g_int_equal);
	
	initBau();
		
	int rank, ntasks;	
	MPI_Init (&argc, &argv);	
	MPI_Comm_rank (MPI_COMM_WORLD, &rank);	
	MPI_Comm_size (MPI_COMM_WORLD, &ntasks);

	if (ntasks != TASKS) {
	  printf("Must specify TASKS= %d. Terminating.\n", TASKS);
	  MPI_Finalize();
	  exit(0);
	} else if(maxRows*maxCols < 300) {
		printf("map must be at least 12x25!");
		MPI_Finalize();
		exit(0);
	}
	
	if(rank == 0) {
		master(ntasks);
	}
	else if(rank == 1) {
		resourceAndFlavorSlave(ntasks);
	}	
	else {		
		slave(rank);
	}
	
	MPI_Finalize();
	return 0;
}